#include <AS.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Str.h>
#include <Cytoplasm/Log.h>
#include <Cytoplasm/Uri.h>
#include <Cytoplasm/Sha.h>

#include <string.h>
#include <stdlib.h>

#include <Matrix.h>

char *
ASUpload(const ParseeConfig *c, Stream *from, unsigned int size, char *mime)
{
    char *size_str, *path, *ret, *user;
    unsigned int i;
    HttpClientContext *ctx;
    HashMap *reply;
    if (!c || !from)
    {
        return NULL;
    }

    size_str = StrInt(size);
    user = StrConcat(4, "@",c->sender_localpart,":",c->server_base);
    path = StrConcat(2,
        "/_matrix/media/v3/upload?user_id=", user
    );
    ctx = ParseeCreateRequest(c, HTTP_POST, path);
    ASAuthenticateRequest(c, ctx);
    if (size)
    {
        HttpRequestHeader(ctx, "Content-Length", size_str);
    }
    if (mime)
    {
        HttpRequestHeader(ctx, "Content-Type", mime);
    }
    HttpRequestSendHeaders(ctx);

    for (i = 0; i < size; i++)
    {
        int ch = StreamGetc(from);
        if (ch == EOF)
        {
            break;
        }
        StreamPutc(HttpClientStream(ctx), ch);
    }
    HttpRequestSend(ctx);
    reply = JsonDecode(HttpClientStream(ctx));
    ret = StrDuplicate(
        JsonValueAsString(HashMapGet(reply, "content_uri"))
    );
    if (!ret)
    {
        JsonEncode(reply, StreamStdout(), JSON_PRETTY);
        StreamFlush(StreamStdout());
    }
    HttpClientContextFree(ctx);
    JsonFree(reply);
    Free(size_str);
    Free(path);
    Free(user);

    return ret;
}
char *
ASReupload(const ParseeConfig *c, char *from, char **mime)
{
    Uri *uri;
    HttpClientContext *ctx;
    unsigned short port;
    int size = 0, flags = HTTP_FLAG_NONE;
    char *ret, *content_len;

    if (!c || !from)
    {
        return NULL;
    }

    uri = UriParse(from);
    if (!uri)
    {
        return NULL;
    }
    if (uri->port)
    {
        port = uri->port;
    }
    else if (StrEquals(uri->proto, "https"))
    {
        port = 443;
    }
    else
    {
        port = 80;
    }
    
    if (StrEquals(uri->proto, "https"))
    {
        flags |= HTTP_FLAG_TLS;
    }

    ctx = HttpRequest(
        HTTP_GET, flags, port, uri->host, uri->path
    );
    HttpRequestSendHeaders(ctx);
    HttpRequestSend(ctx);

    if (mime)
    {
        *mime = HashMapGet(HttpResponseHeaders(ctx), "content-type");
        *mime = StrDuplicate(*mime);
    }

    content_len = HashMapGet(HttpResponseHeaders(ctx), "content-length");
    if (content_len)
    {
        size = strtol(content_len, NULL, 10);
    }
    ret = ASUpload(c, HttpClientStream(ctx), size, mime ? *mime : NULL);

    HttpClientContextFree(ctx);
    UriFree(uri);
    return ret;
}

bool
ASGetMIMESHA(const ParseeConfig *c, char *mxc, char **mime, char **sha)
{
    HttpClientContext *cctx;
    Stream *stream;
    Stream *fake;
    Uri *uri;
    char *path, *buf = NULL;
    unsigned char *sha1;
    size_t len;
    if (!c || !mxc || !mime || !sha)
    {
        return false;
    }
    *mime = NULL;
    *sha = NULL;

    if (!(uri = UriParse(mxc)) || !StrEquals(uri->proto, "mxc"))
    {
        return false;
    }

    path = StrConcat(3, "/_matrix/media/v3/download/", uri->host, uri->path);
    cctx = ParseeCreateRequest(c, HTTP_GET, path);
    ASAuthenticateRequest(c, cctx);
    HttpRequestSendHeaders(cctx);
    HttpRequestSend(cctx);

    *mime = StrDuplicate(
        HashMapGet(HttpResponseHeaders(cctx), "content-type")
    );
    stream = HttpClientStream(cctx);
    fake = StreamFile(open_memstream(&buf, &len));
    StreamCopy(stream, fake);
    StreamClose(fake);

    sha1 = Sha1Raw((unsigned char *) buf, len);
    free(buf);
    *sha = ShaToHex(sha1, HASH_SHA1);
    Free(sha1);

    HttpClientContextFree(cctx);
    UriFree(uri);
    Free(path);
    return true;
}
bool
ASGrab(const ParseeConfig *c, char *mxc, char **mime, char **out, size_t *len)
{
    HttpClientContext *cctx;
    Stream *stream;
    Stream *fake;
    Uri *uri;
    char *path, *buf = NULL;
    if (!c || !mxc || !mime || !out || !len)
    {
        return false;
    }
    *mime = NULL;
    *out = NULL;
    *len = 0;

    if (!(uri = UriParse(mxc)) || !StrEquals(uri->proto, "mxc"))
    {
        return false;
    }

    path = StrConcat(3, "/_matrix/media/v3/download/", uri->host, uri->path);
    cctx = ParseeCreateRequest(c, HTTP_GET, path);
    ASAuthenticateRequest(c, cctx);
    HttpRequestSendHeaders(cctx);
    HttpRequestSend(cctx);

    *mime = StrDuplicate(
        HashMapGet(HttpResponseHeaders(cctx), "content-type")
    );
    stream = HttpClientStream(cctx);
    fake = StreamFile(open_memstream(&buf, len));
    StreamCopy(stream, fake);
    StreamClose(fake);

    *out = Malloc(*len);
    memcpy(*out, buf, *len);
    free(buf);

    HttpClientContextFree(cctx);
    UriFree(uri);
    Free(path);
    return true;
}
