#include <Parsee.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Json.h>
#include <Cytoplasm/Util.h>
#include <Cytoplasm/Str.h>
#include <Cytoplasm/Log.h>

#include <stdlib.h>
#include <string.h>

static ParseeConfig *config = NULL;

void
ParseeConfigInit(void)
{
    if (config)
    {
        return;
    }
    Log(LOG_ERR, "No config file found.");
    Log(LOG_ERR, "Please use parsee-config to initialise %s.", NAME);
}
void
ParseeConfigLoad(char *conf)
{
    Stream *stream;
    HashMap *json;
    if (config)
    {
        return;
    }
    stream = StreamOpen(conf ? conf : "parsee.json", "r");
    if (!stream)
    {
        return;
    }
    json = JsonDecode(stream);
    if (!json)
    {
        Log(LOG_ERR, "Could not parse config JSON");
        StreamClose(stream);
        return;
    }

    config = Malloc(sizeof(*config));
#define CopyToStr(to, str)  config->to = StrDuplicate( \
                                JsonValueAsString(HashMapGet(json, str)) \
                            )
#define CopyToInt(to, str)  config->to = (int) ( \
                                JsonValueAsInteger(HashMapGet(json, str)) \
                            )
#define CopyToBool(to, str)  config->to = (int) ( \
                                JsonValueAsBoolean(HashMapGet(json, str)) \
                            )

    config->http_threads = 8;
    config->xmpp_threads = 8;
    
    CopyToStr(as_token, "as_token");
    CopyToStr(hs_token, "hs_token");
    CopyToStr(sender_localpart, "sender");
    CopyToStr(namespace_base, "namespace");
    CopyToStr(listen_as, "listen_as");
    CopyToInt(port, "port");

    CopyToStr(server_base, "hs_base");
    CopyToStr(homeserver_host, "hs_host");
    CopyToInt(homeserver_port, "hs_port");
    CopyToBool(homeserver_tls, "hs_tls");
    if (!HashMapGet(json, "hs_tls"))
    {
        config->homeserver_tls = true;
    }
    CopyToBool(accept_pings, "accept_pings");

    CopyToInt(component_port, "component_port");
    CopyToStr(component_addr, "component_addr");
    CopyToStr(component_host, "component_host");
    CopyToStr(shared_comp_secret, "shared_secret");
    CopyToInt(max_stanza_size, "max_stanza_size");
    if (!config->max_stanza_size)
    {
        /* Standard XMPP "minimum" maximum */
        config->max_stanza_size = 10000;
    }

    CopyToStr(media_base, "media_base");

    CopyToStr(db_path, "db");
    CopyToInt(db_size, "db_size");

    JsonFree(json);
    StreamClose(stream);
}

void
ParseeSetThreads(int xmpp, int http)
{
    if (!config)
    {
        Achievement("THREAD COUNT REQUEST WITHOUT CONFIG", true);
        return;
    }
    config->http_threads = http;
    config->xmpp_threads = xmpp;
}

void
ParseeExportConfigYAML(Stream *stream)
{
    if (!stream || !config)
    {
        Achievement("YAML EXPORT REQUEST WITHOUT CONFIG", true);
        return;
    }
    StreamPrintf(stream, "# Autogenerated YAML AS entry for %s\n", NAME);
    StreamPrintf(stream, "# This file should be *moved* to your\n");
    StreamPrintf(stream, "# Matrix homeserver's config directory\n");
    StreamPrintf(stream, "# (as for Conduit, y'all really need a better system)\n");
    StreamPrintf(stream, "\n");
    StreamPrintf(stream, "id: \"Parsee XMPP\"\n");
    StreamPrintf(stream, "url: \"http://%s:%d/\"\n", config->listen_as, config->port);
    StreamPrintf(stream, "as_token: \"%s\"\n", config->as_token);
    StreamPrintf(stream, "hs_token: \"%s\"\n", config->hs_token);
    StreamPrintf(stream, "sender_localpart: \"%s\"\n", config->sender_localpart);
    StreamPrintf(stream, "protocols: [\"xmpp\", \"jabber\"]\n");
    StreamPrintf(stream, "receive_ephemeral: true\n"); /* TODO: Actually use that field */
    StreamPrintf(stream, "\n");
    StreamPrintf(stream, "namespaces: \n");
    StreamPrintf(stream, "    users:\n");
    StreamPrintf(stream, "        - exclusive: true\n");
    StreamPrintf(stream, "          regex: \"@%s_.*\"\n", config->namespace_base);
    StreamPrintf(stream, "    aliases:\n");
    StreamPrintf(stream, "        - exclusive: true\n");
    StreamPrintf(stream, "          regex: \"#%s_.*\"\n", config->namespace_base);
    StreamFlush(stream);
}

void
ParseeConfigFree(void)
{
    if (!config)
    {
        return;
    }
    Free(config->component_host);
    Free(config->component_addr);
    Free(config->shared_comp_secret);
    Free(config->db_path);
    Free(config->homeserver_host);
    
    Free(config->as_token);
    Free(config->hs_token);
    Free(config->server_base);
    Free(config->sender_localpart);
    Free(config->namespace_base);
    Free(config->media_base);
    Free(config->listen_as);
    Free(config);
}
const ParseeConfig *
ParseeConfigGet(void)
{
    return (const ParseeConfig *) config;
}
