#include <Parsee.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Json.h>
#include <Cytoplasm/Util.h>
#include <Cytoplasm/Log.h>
#include <Cytoplasm/Str.h>

#include <stdlib.h>
#include <string.h>

#include <Routes.h>
#include <AS.h>

ParseeData *
ParseeInitData(XMPPComponent *comp)
{
    char *version;
    ParseeData *data;
    DbRef *ref;
    if (!ParseeConfigGet())
    {
        return NULL;
    }

    data = Malloc(sizeof(*data));
    data->config = ParseeConfigGet();
    data->router = HttpRouterCreate();
    data->jabber = comp;
    data->muc = CreateMUCServer(data);
    data->handler = CommandCreateRouter();

    data->oid_servers = HashMapCreate();
    pthread_mutex_init(&data->oidl, NULL);

    data->halted = false;
    pthread_mutex_init(&data->halt_lock, NULL);

    if (data->config->db_size)
    {
        data->db = DbOpenLMDB(data->config->db_path, data->config->db_size);
    }
    if (!data->db)
    {
        Log(LOG_WARNING, "LMDB doesn't seem to be setup.");
        if (data->config->db_size)
        {
            Log(LOG_WARNING, "Falling back to flat-file.");
        }
        data->db = DbOpen(data->config->db_path, 0);
    }

    if (!(ref = DbLock(data->db, 1, "info")))
    {
        char *id = StrRandom(64);
        ref = DbCreate(data->db, 1, "info");
        HashMapSet(DbJson(ref), "identifier", JsonValueString(id));
        HashMapSet(DbJson(ref), "version", JsonValueString(VERSION));
        Free(id);
    }

    version = GrabString(DbJson(ref), 1, "version");
    if (version && !ParseeIsCompatible(VERSION, version))
    {
        Log(LOG_WARNING, "Version mismatch(curr=%s db=%s).", VERSION, version);
        Log(LOG_WARNING, "Yeah. You may want to _not_ do that.");
        Log(LOG_WARNING, "(Parsee still needs an upgradepath mechanism.)");

        DbUnlock(data->db, ref);
        DbClose(data->db);

        HashMapFree(data->oid_servers);
        pthread_mutex_destroy(&data->oidl);

        XMPPEndCompStream(data->jabber);

        HttpRouterFree(data->router);
        CommandFreeRouter(data->handler);

        Free(data);
        return NULL;
    }

    data->id = StrDuplicate(GrabString(DbJson(ref), 1, "identifier"));
    DbUnlock(data->db, ref);

#define X_ROUTE(path, func) do {\
                                if (!HttpRouterAdd(data->router, path, func))\
                                {\
                                    Log(LOG_ERR, "Can't register %s", path);\
                                }\
                            }\
                            while (0);
    ROUTES
#undef X_ROUTE
#define X_COMMAND(path,n,d) CommandAddCommand(data->handler, path, n);
    COMMANDS
#undef X_COMMAND
    return data;
}
void
ParseeFreeData(ParseeData *data)
{
    char *entity;
    XMLElement *disco;
    if (!data)
    {
        return;
    }

    while (HashMapIterate(data->oid_servers, &entity, (void **) &disco))
    {
        XMLFreeElement(disco);
    }
    HashMapFree(data->oid_servers);
    pthread_mutex_destroy(&data->oidl);
    pthread_mutex_destroy(&data->halt_lock);
    Free(data->id);
    XMPPEndCompStream(data->jabber);
    FreeMUCServer(data->muc);
    DbClose(data->db);
    HttpRouterFree(data->router);
    CommandFreeRouter(data->handler);
    Free(data);
}

void 
ParseeCleanup(void *datp)
{
    ParseeData *data = datp;
    Array *chats;
    char *chat;
    size_t i;
    uint64_t ts = UtilTsMillis();

    chats = DbList(data->db, 1, "chats");

    for (i = 0; i < ArraySize(chats); i++)
    {
        DbRef *ref;
        char *stanza, *event, *id;
        Array *to_delete;
        size_t j;

        chat = ArrayGet(chats, i);

#define CleanupField(field, timeout, threshold) do \
    { \
        Array *field##_list = DbList(data->db, 3, "chats", chat, #field"s"); \
        size_t cleaned = 0; \
        to_delete = ArrayCreate(); \
        for (j = 0; j < ArraySize(field##_list); j++) \
        { \
            char *f = ArrayGet(field##_list, j); \
            ref = DbLockIntent(data->db, \
                DB_HINT_READONLY, \
                4, "chats", chat, #field"s", f \
            ); \
            HashMap *obj = DbJson(ref); \
            uint64_t age = JsonValueAsInteger(HashMapGet(obj, "age")); \
            uint64_t dur = ts - age; \
 \
            if ((dur > (timeout))) \
            { \
                ArrayAdd(to_delete, StrDuplicate(f)); \
                cleaned++; \
            } \
            DbUnlock(data->db, ref); \
        } \
        DbListFree(field##_list); \
 \
        for (j = 0; j < ArraySize(to_delete); j++) \
        { \
            field = ArrayGet(to_delete, j); \
            if (cleaned > threshold) \
            { \
                DbDelete(data->db, 4, "chats", chat, #field"s", field); \
            } \
            Free(field); \
        } \
        ArrayFree(to_delete); \
    } \
    while (0)

        /* TODO: Custom retention period for any 1.0 */
        CleanupField(stanza,    30 MINUTES, 500);
        CleanupField(event,     30 MINUTES, 500);
        CleanupField(id,        30 MINUTES, 500);
        
        /* TODO: Also cleanup user cache information */
#undef CleanupField
    }
    DbListFree(chats);
    /* TODO */
    chats = DbList(data->db, 1, "rooms");

    for (i = 0; i < ArraySize(chats); i++)
    {
        DbRef *ref;
        HashMap *json, *stanzas, *events, *ids;
        char *stanza, *event, *id;
        JsonValue *val;
        Array *to_delete;
        size_t j;

        chat = ArrayGet(chats, i);
        ref = DbLock(data->db, 3, "rooms", chat, "data");
        json = DbJson(ref);

#define CleanupField(field, timeout, threshold) do \
    { \
        size_t cleaned = 0; \
        field##s = JsonValueAsObject(HashMapGet(json, #field"s")); \
        to_delete = ArrayCreate(); \
        while (HashMapIterate(field##s, &field, (void **) &val)) \
        { \
            HashMap *obj = JsonValueAsObject(val); \
            uint64_t age = JsonValueAsInteger(HashMapGet(obj, "age")); \
            uint64_t dur = ts - age; \
 \
            if ((dur > (timeout))) \
            { \
                ArrayAdd(to_delete, StrDuplicate(field)); \
                cleaned++; \
            } \
        } \
 \
        for (j = 0; j < ArraySize(to_delete); j++) \
        { \
            field = ArrayGet(to_delete, j); \
            if (cleaned > threshold) \
            { \
                JsonValueFree(HashMapDelete(field##s, field)); \
            } \
            Free(field); \
        } \
        ArrayFree(to_delete); \
    } \
    while (0)

        CleanupField(stanza,    3 HOURS, 500);
        CleanupField(event,     3 HOURS, 500);
        CleanupField(id,        3 HOURS, 500);

        DbUnlock(data->db, ref);
    }
    DbListFree(chats);
}

void
ParseePushDMStanza(ParseeData *data, char *room_id, char *stanza_id, char *id, char *ev, char *sender)
{
    DbRef *ref;
    HashMap *j;
    HashMap *stanzas, *obj, *events, *ids;
    bool new_stanzas = false, new_events = false;
    bool new_ids = false;
    uint64_t age = UtilTsMillis();
    if (!data || !room_id || !ev || !sender)
    {
        return;
    }

    ref = DbLock(data->db, 3, "rooms", room_id, "data");
    if (!ref)
    {
        return;
    }
    j = DbJson(ref);


    if (stanza_id)
    {
        stanzas = JsonValueAsObject(HashMapGet(j, "stanzas"));
        if (!stanzas)
        {
            stanzas = HashMapCreate();
            new_stanzas = true;
        }

        obj = HashMapCreate();
        HashMapSet(obj, "age", JsonValueInteger(age));
        HashMapSet(obj, "event", JsonValueString(ev));
        JsonValueFree(HashMapSet(stanzas, stanza_id, JsonValueObject(obj)));

        if (new_stanzas)
        {
            HashMapSet(j, "stanzas", JsonValueObject(stanzas));
        }
    }

    events = JsonValueAsObject(HashMapGet(j, "events"));
    if (!events)
    {
        events = HashMapCreate();
        new_events = true;
    }
    obj = HashMapCreate();
    HashMapSet(obj, "stanza", JsonValueString(stanza_id));
    HashMapSet(obj, "origin", JsonValueString(id));
    HashMapSet(obj, "sender", JsonValueString(sender));
    HashMapSet(obj, "age", JsonValueInteger(age));
    JsonValueFree(HashMapSet(events, ev, JsonValueObject(obj)));
    if (new_events)
    {
        HashMapSet(j, "events", JsonValueObject(events));
    }

    if (id)
    {
        ids = JsonValueAsObject(HashMapGet(j, "ids"));
        if (!ids)
        {
            ids = HashMapCreate();
            new_ids = true;
        }
        obj = HashMapCreate();
        HashMapSet(obj, "stanza", JsonValueString(stanza_id));
        HashMapSet(obj, "event", JsonValueString(ev));
        HashMapSet(obj, "sender", JsonValueString(sender));
        HashMapSet(obj, "age", JsonValueInteger(age));
        JsonValueFree(HashMapSet(ids, id, JsonValueObject(obj)));
        if (new_ids)
        {
            HashMapSet(j, "ids", JsonValueObject(ids));
        }
    }
    DbUnlock(data->db, ref);
}
void
ParseePushStanza(ParseeData *data, char *chat_id, char *stanza_id, char *id, char *ev, char *sender)
{
    DbRef *ref;
    HashMap *j;
    uint64_t age = UtilTsMillis();
    if (!data || !chat_id || !stanza_id || !ev || !sender)
    {
        return;
    }

    /* TODO */
    {
        ref = DbLock(data->db, 2, "chats", chat_id);
        j = DbJson(ref);
        if (j)
        {
            JsonValueFree(HashMapSet(j, "ts", JsonValueInteger(age)));
        }
        DbUnlock(data->db, ref);
    }
    {
        ref = DbCreate(data->db, 4, "chats", chat_id, "stanzas", stanza_id);
        j = DbJson(ref);
        if (!ref)
        {
            return;
        }

        HashMapSet(j, "age", JsonValueInteger(age));
        HashMapSet(j, "event", JsonValueString(ev));
        DbUnlock(data->db, ref);
    }

    {
        ref = DbCreate(data->db, 4, "chats", chat_id, "events", ev);
        j = DbJson(ref);
        if (!ref)
        {
            return;
        }

        HashMapSet(j, "event", JsonValueString(ev));
        HashMapSet(j, "stanza", JsonValueString(stanza_id));
        HashMapSet(j, "origin", JsonValueString(id));
        HashMapSet(j, "sender", JsonValueString(sender));
        HashMapSet(j, "age", JsonValueInteger(age));
        DbUnlock(data->db, ref);
    }

    if (id)
    {
        ref = DbCreate(data->db, 4, "chats", chat_id, "ids", id);
        j = DbJson(ref);
        if (!ref)
        {
            return;
        }
        HashMapSet(j, "stanza", JsonValueString(stanza_id));
        HashMapSet(j, "event", JsonValueString(ev));
        HashMapSet(j, "sender", JsonValueString(sender));
        HashMapSet(j, "age", JsonValueInteger(age));
        DbUnlock(data->db, ref);
    }
}

bool
ParseeVerifyDMStanza(ParseeData *data, char *room_id, char *id)
{
    DbRef *ref = NULL;
    HashMap *j = NULL;
    HashMap *stanzas = NULL;
    bool ret = true;
    if (!data || !room_id || !id)
    {
        return true;
    }

    ref = DbLockIntent(data->db, 
        DB_HINT_READONLY,
        3, "rooms", room_id, "data"
    );
    j = DbJson(ref);
    if (!ref)
    {
        goto end;
    }

    stanzas = JsonValueAsObject(HashMapGet(j, "ids"));
    if (!stanzas)
    {
        goto end;
    }

    ret = !HashMapGet(stanzas, id);
end:
    DbUnlock(data->db, ref);
    return ret;
}
bool
ParseeVerifyStanza(ParseeData *data, char *chat_id, char *stanza_id)
{
    bool ret = true;
    if (!data || !chat_id || !stanza_id)
    {
        return true;
    }

    ret = !DbExists(data->db, 4, "chats", chat_id, "stanzas", stanza_id);
    return ret;
}
char *
ParseeEventFromID(ParseeData *data, char *chat_id, char *id)
{
    DbRef *ref = NULL;
    HashMap *j = NULL;
    char *ret = NULL;
    if (!data || !chat_id || !id)
    {
        return NULL;
    }

    ref = DbLock(data->db, 4, "chats", chat_id, "ids", id);
    j = DbJson(ref);
    if (!ref)
    {
        goto end;
    }

    ret = JsonValueAsString(JsonGet(j, 1, "event"));
    ret = StrDuplicate(ret);
end:
    DbUnlock(data->db, ref);
    return ret;
}
char *
ParseeDMEventFromID(ParseeData *data, char *room_id, char *id)
{
    DbRef *ref = NULL;
    HashMap *j = NULL;
    HashMap *ids = NULL;
    char *ret = NULL;
    if (!data || !room_id || !id)
    {
        return NULL;
    }

    ref = DbLock(data->db, 3, "rooms", room_id, "data");
    j = DbJson(ref);
    if (!ref)
    {
        goto end;
    }

    ids = JsonValueAsObject(HashMapGet(j, "ids"));
    if (!ids)
    {
        goto end;
    }

    ret = JsonValueAsString(JsonGet(ids, 2, id, "event"));
    ret = StrDuplicate(ret);
end:
    DbUnlock(data->db, ref);
    return ret;
}

char *
ParseeEventFromSID(ParseeData *data, char *chat_id, char *id)
{
    DbRef *ref = NULL;
    HashMap *j = NULL;
    char *ret = NULL;
    if (!data || !chat_id || !id)
    {
        return NULL;
    }

    ref = DbLock(data->db, 4, "chats", chat_id, "stanzas", id);
    j = DbJson(ref);
    if (!ref)
    {
        goto end;
    }

    ret = JsonValueAsString(JsonGet(j, 1, "event"));
    ret = StrDuplicate(ret);
end:
    DbUnlock(data->db, ref);
    return ret;
}
char *
ParseeDMEventFromSID(ParseeData *data, char *room_id, char *id)
{
    DbRef *ref = NULL;
    HashMap *j = NULL;
    HashMap *ids = NULL;
    char *ret = NULL;
    if (!data || !room_id || !id)
    {
        return NULL;
    }

    ref = DbLock(data->db, 3, "rooms", room_id, "data");
    j = DbJson(ref);
    if (!ref)
    {
        goto end;
    }

    ids = JsonValueAsObject(HashMapGet(j, "stanzas"));
    if (!ids)
    {
        goto end;
    }

    ret = JsonValueAsString(JsonGet(ids, 2, id, "event"));
    ret = StrDuplicate(ret);
end:
    DbUnlock(data->db, ref);
    return ret;

}

void 
ParseeUnlinkRoom(ParseeData *data, char *chat_id)
{
    char *muc, *room;
    DbRef *ref;
    if (!data || !chat_id)
    {
        return;
    }

    muc = ParseeGetMUCID(data, chat_id);
    room = ParseeGetRoomID(data, chat_id);
    if (!muc || !room)
    {
        Free(muc);
        Free(room);
        return;
    }

    ref = DbLock(data->db, 1, "chats");
    JsonValueFree(HashMapDelete(
        GrabObject(DbJson(ref), 1, "rooms"), 
        room
    ));
    JsonValueFree(HashMapDelete(
        GrabObject(DbJson(ref), 1, "mucs"), 
        muc
    ));
    DbUnlock(data->db, ref);
    DbDelete(data->db, 2, "chats", chat_id);

    Free(muc);
    Free(room);
}
bool 
ParseeIsMUCWhitelisted(ParseeData *data, char *muc)
{
    char *server, *serv_start, *postserv;
    DbRef *ref;
    bool ret;
    if (!data || !muc)
    {
        return false;
    }

    if (!DbExists(data->db, 1, "whitelist"))
    {
        return true;
    }

    serv_start = strchr(muc, '@');
    serv_start = serv_start ? serv_start : muc;
    server = StrDuplicate(serv_start + 1);
    postserv = server ? strchr(server, '/') : NULL;
    if (postserv)   /* GCC doesn't know strchr is pure. */
    {
        *postserv = '\0';
    }
    
    ref = DbLockIntent(data->db, 
        DB_HINT_READONLY,
        1, "whitelist"
    );
    ret = HashMapGet(DbJson(ref), server);
    DbUnlock(data->db, ref);
    Free(server);

    return ret;
}

extern HashMap *
ParseeGetChatSettings(ParseeData *data, char *chat)
{
    HashMap *ret, *json;
    DbRef *ref;

    char *key;
    JsonValue *value;
    if (!data || !chat)
    {
        return NULL;
    }

    ref = DbLockIntent(data->db, DB_HINT_READONLY, 
        3, "chats", chat, "settings"
    );
    json = DbJson(ref);
    if (!ref)
    {
        return HashMapCreate();
    }

    ret = HashMapCreate();
    while (HashMapIterate(json, &key, (void **) &value))
    {
        char *str = JsonValueAsString(value);
        HashMapSet(ret, key, StrDuplicate(str));
    }
    DbUnlock(data->db, ref);
    return ret;
}
void
ParseeFreeChatSettings(HashMap *settings)
{
    char *key;
    void *val;
    if (!settings)
    {
        return;
    }
    while (HashMapIterate(settings, &key, &val))
    {
        Free(val);
    }
    HashMapFree(settings);
}
char * 
ParseeGetChatSetting(ParseeData *data, char *chat, char *key)
{
    HashMap *map;
    char *ret;
    if (!data || !chat || !key)
    {
        return NULL;
    }

    map = ParseeGetChatSettings(data, chat);
    ret = StrDuplicate(HashMapGet(map, key));
    ParseeFreeChatSettings(map);

    return ret;
}
void 
ParseeSetChatSetting(ParseeData *data, char *chat, char *key, char *val)
{
    DbRef *ref;
    HashMap *json;
    if (!data || !chat || !key || !val)
    {
        return;
    }

    ref = DbLockIntent(data->db, DB_HINT_WRITE, 
        3, "chats", chat, "settings"
    );
    if (!ref)
    {
        ref = DbCreate(data->db, 3, "chats", chat, "settings");
    }
    json = DbJson(ref);

    JsonValueFree(HashMapSet(json, key, JsonValueString(val)));

    DbUnlock(data->db, ref);
    return;
}
bool
ParseeIsMediaEnabled(ParseeData *data, char *chat_id)
{
    char *value;
    bool ret;
    if (!data || !chat_id)
    {
        return false;
    }

    ret = !StrEquals(
        (value = ParseeGetChatSetting(data, chat_id, "p.media.enabled")),
        "false"
    );
    Free(value);

    return ret;
}
