#include <XMPP.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Util.h>
#include <Cytoplasm/Str.h>
#include <Cytoplasm/Log.h>

#include <Parsee.h>
#include <XML.h>

#include "XMPPThread/internal.h"

bool
XMPPQueryMUC(XMPPComponent *jabber, char *muc, MUCInfo *out)
{
    XMLElement *iq_query, *query;
    char *uuid, *from;
    if (!jabber || !muc)
    {
        return false;
    }

    iq_query = XMLCreateTag("iq");
    query = XMLCreateTag("query");

    XMLAddAttr(iq_query, "from",(from = StrConcat(2,"parsee@",jabber->host)));
    XMLAddAttr(iq_query, "to", muc);
    XMLAddAttr(iq_query, "id", (uuid = StrRandom(8)));
    XMLAddAttr(iq_query, "type", "get");

    XMLAddAttr(query, "xmlns", "http://jabber.org/protocol/disco#info");

    XMLAddChild(iq_query, query);
    Free(from);

    {
        XMLElement *identity;
        XMPPSendStanza(jabber, iq_query, 10000);
        XMLFreeElement(iq_query);

        /* Except an IQ reply. 10 seconds of timeout is pretty 
         * generous, if you ask me. */
        iq_query = ParseeAwaitStanza(uuid, 10 SECONDS);
        Free(uuid);
        if (!iq_query || !StrEquals(iq_query->name, "iq"))
        {
            Log(LOG_ERR, "Didn't receive an <iq> stanza");
            XMLFreeElement(iq_query);
            return false;
        }
        query = XMLookForUnique(iq_query, "query");
        identity = XMLookForUnique(query, "identity");

        if (!identity || 
            !StrEquals(HashMapGet(identity->attrs, "category"),
                "conference"))
        {
            XMLFreeElement(iq_query);
            Log(LOG_DEBUG, "MUC INFO ERROR");
            Log(LOG_DEBUG, 
                "identityp=%p category=%s", identity, 
                identity ? HashMapGet(identity->attrs, "category") : NULL
            );
            return false;
        }

        /* We found a MUC! */
        if (out)
        {
            out->exists = true;
            out->jabber = jabber;
            out->iq_reply = iq_query;
        }
        else
        {
            XMLFreeElement(iq_query);
        }
    }
    return true;
}

void
XMPPFreeMUCInfo(MUCInfo info)
{
    if (!info.exists)
    {
        return;
    }

    XMLFreeElement(info.iq_reply);

    info.exists = false;
}

char *
XMPPGetMUCName(MUCInfo info)
{
    XMLElement *query, *identity;
    char *name;
    if (!info.exists)
    {
        return NULL;
    }
    query = XMLookForUnique(info.iq_reply, "query");
    identity = XMLookForUnique(query, "identity");

    name = StrDuplicate(HashMapGet(identity->attrs, "name"));

    return name;
}
void
XMPPRequestVoice(XMPPComponent *jabber, char *from, char *muc)
{
    XMLElement *stanza;
    char *identifier;
    if (!jabber || !from || !muc)
    {
        return;
    }

    stanza = XMLCreateTag("message");
    XMLAddAttr(stanza, "id", (identifier = StrRandom(32)));
    XMLAddAttr(stanza, "from", from);
    XMLAddAttr(stanza, "to", muc);
    {
        XMLElement *x = XMLCreateTag("x");
        XMLAddAttr(x, "xmlns", "jabber:x:data");
        XMLAddAttr(x, "type", "submit");
        {
            XMLElement *field;

            field = XMLCreateTag("field");
            XMLAddAttr(field, "var", "FORM_TYPE");
            {
                XMLElement *value = XMLCreateTag("value");
                {
                    XMLElement *data = XMLCreateText(
                        "http://jabber.org/protocol/muc#request"
                    );
                    XMLAddChild(value, data);
                }
                XMLAddChild(field, value);
            }
            XMLAddChild(x, field);

            field = XMLCreateTag("field");
            XMLAddAttr(field, "var", "muc#role");
            XMLAddAttr(field, "type", "list-single");
            XMLAddAttr(field, "label", "Requested role");
            {
                XMLElement *value = XMLCreateTag("value");
                {
                    XMLElement *data = XMLCreateText(
                        "participant"
                    );
                    XMLAddChild(value, data);
                }
                XMLAddChild(field, value);
            }
            XMLAddChild(x, field);
        }
        XMLAddChild(stanza, x);
    }
    XMPPSendStanza(jabber, stanza, 10000);

    XMLFreeElement(stanza);
    Free(identifier);
}
bool
XMPPJoinMUC(XMPPComponent *comp, char *fr, char *muc, char *hash, int time, bool ret)
{
    XMLElement *presence, *x, *reply, *history, *photo;
    IQFeatures *features;
    char *from, *id, *stime = "3600";
    if (!comp || !fr || !muc)
    {
        return false;
    }

    presence = XMLCreateTag("presence");
    x = XMLCreateTag("x");
    XMLAddAttr(presence, "from", (from = StrConcat(3, fr, "@", comp->host)));
    XMLAddAttr(presence, "to", muc);
    XMLAddAttr(presence, "id", (id = StrRandom(8)));
    XMLAddAttr(x, "xmlns", "http://jabber.org/protocol/muc");
    history = XMLCreateTag("history");

    if (time > 0)
    {
        stime = StrInt(time);
    }
    XMLAddAttr(history, "seconds", stime);
    if (time > 0)
    {
        Free(stime);
        stime = NULL;
    }
    XMLAddChild(x, history);

    XMLAddChild(presence, x);
    features = LookupJIDFeatures(from);
#define IdentitySimple(cat, Type, Name) AddIQIdentity(features, cat, NULL, Type, Name);
    IQ_IDENTITY
#undef IdentitySimple
#define AdvertiseSimple(feature) AdvertiseIQFeature(features, feature);
    IQ_ADVERT
#undef AdvertiseSimple
    XMPPAnnotatePresence(presence, features);
    FreeIQFeatures(features);

    if (hash)
    {
        x = XMLCreateTag("x");
        XMLAddAttr(x, "xmlns", "vcard-temp:x:update");
        photo = XMLCreateTag("photo");
        XMLAddChild(photo, XMLCreateText(hash));
        XMLAddChild(x, photo);
        XMLAddChild(presence, x);
    }

    XMPPSendStanza(comp, presence, 10000);
    XMLFreeElement(presence);
    Free(from);

    if (ret && (reply = ParseeAwaitStanza(id, 500)))
    {
        bool exit_code = true;

        if (XMPPHasError(reply, "conflict"))
        {
            exit_code = false;
        }

        XMLFreeElement(reply);
        Free(id);
        return exit_code;
    }
    Free(id);
    return true;
}
void
XMPPLeaveMUC(XMPPComponent *comp, char *fr, char *muc, char *reason)
{
    XMLElement *presence;
    IQFeatures *features;
    char *from, *id;
    if (!comp || !fr || !muc)
    {
        return;
    }

    presence = XMLCreateTag("presence");
    XMLAddAttr(presence, "from", (from = StrConcat(3, fr, "@", comp->host)));
    XMLAddAttr(presence, "to", muc);
    XMLAddAttr(presence, "id", (id = StrRandom(8)));
    XMLAddAttr(presence, "type", "unavailable");

    if (reason)
    {
        XMLElement *status = XMLCreateTag("status");
        XMLElement *string = XMLCreateText(reason);

        XMLAddChild(status, string);
        XMLAddChild(presence, status);
    }

    features = LookupJIDFeatures(from);
#define IdentitySimple(cat, Type, Name) AddIQIdentity(features, cat, NULL, Type, Name);
    IQ_IDENTITY
#undef IdentitySimple
#define AdvertiseSimple(feature) AdvertiseIQFeature(features, feature);
    IQ_ADVERT
#undef AdvertiseSimple
    XMPPAnnotatePresence(presence, features);
    FreeIQFeatures(features);


    XMPPSendStanza(comp, presence, 10000);

    XMLFreeElement(presence);
    Free(from);
    Free(id);
}
