#include <XMPP.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Str.h>
#include <Cytoplasm/Log.h>

#include <Parsee.h>
#include <XML.h>

#include <stdlib.h>


void
XMPPRetract(XMPPComponent *comp, char *fr, char *to, char *type, char *redact)
{
    XMLElement *message;
    char *ident = StrRandom(32), *from;
    if (!comp || !fr || !to || !type || !redact)
    {
        Free(ident);
        return;
    }

    message = XMLCreateTag("message");
    XMLAddAttr(message, "from", (from = StrConcat(3, fr, "@", comp->host)));
    XMLAddAttr(message, "to", to);
    XMLAddAttr(message, "type", type);
    XMLAddAttr(message, "id", ident);

    {
        XMLElement *apply_to, *retract, *body, *txt, *fallback;

        apply_to = XMLCreateTag("apply-to");
        XMLAddAttr(apply_to, "xmlns", "urn:xmpp:fasten:0");
        XMLAddAttr(apply_to, "id", redact);
        {
            retract = XMLCreateTag("retract");
            XMLAddAttr(retract, "xmlns", "urn:xmpp:message-retract:0");
            XMLAddChild(apply_to, retract);
        }
        XMLAddChild(message, apply_to);

        retract = XMLCreateTag("retract");
        XMLAddAttr(retract, "xmlns", "urn:xmpp:message-retract:1");
        XMLAddAttr(retract, "id", redact);
        XMLAddChild(message, retract);

        fallback = XMLCreateTag("fallback");
        XMLAddAttr(fallback, "xmlns", "urn:xmpp:fallback:0");
        XMLAddAttr(fallback, "for", "urn:xmpp:message-retract:0");
        XMLAddChild(message, fallback);

        fallback = XMLCreateTag("fallback");
        XMLAddAttr(fallback, "xmlns", "urn:xmpp:fallback:0");
        XMLAddAttr(fallback, "for", "urn:xmpp:message-retract:1");
        XMLAddChild(message, fallback);

        body = XMLCreateTag("body");
        txt = XMLCreateText(
            "This person attempted to retract a previous message, "
            "but your client does not yet support the 0.4.1 version of "
            "XEP-0424.\n\n"

            "(_Parsee is jealous of being able to read such messages._)"
        );
        XMLAddChild(body, txt);
        XMLAddChild(message, body);

    }

    XMPPSendStanza(comp, message, 10000);
    XMLFreeElement(message);

    Free(from);
    Free(ident);
}

bool
XMPPIsParseeStanza(XMLElement *stanza)
{
    if (!stanza)
    {
        return false;
    }
    
    /* TODO: Check if the user is a trustworthy Parsee puppet instead of some 
     * guy sending random stanzas */
    return !!XMLookForUnique(stanza, "x-parsee");
}


char *
XMPPGetStanzaID(XMLElement *stanza)
{
    XMLElement *stanza_id;
    if (!stanza)
    {
        return NULL;
    }

    stanza_id = XMLookForTKV(stanza, "stanza-id", "xmlns", "urn:xmpp:sid:0");
    return stanza_id ? HashMapGet(stanza_id->attrs, "id") : NULL;
}
char *
XMPPGetOriginID(XMLElement *stanza)
{
    XMLElement *origin_id;
    if (!stanza)
    {
        return NULL;
    }

    origin_id = XMLookForTKV(stanza, "origin-id", "xmlns", "urn:xmpp:sid:0");
    return origin_id ? HashMapGet(origin_id->attrs, "id") : NULL;
}
char *
XMPPGetReplacedID(XMLElement *stanza)
{
    XMLElement *replace = XMLookForTKV(
        stanza, "replace", 
        "xmlns", "urn:xmpp:message-correct:0"
    );
    return replace ? HashMapGet(replace->attrs, "id") : NULL;
}
char *
XMPPGetRetractedID(XMLElement *stanza)
{
    XMLElement *retract = XMLookForTKV(
        stanza, "retract", 
        "xmlns", "urn:xmpp:message-retract:1"
    );
    char *id = retract ? HashMapGet(retract->attrs, "id") : NULL;
    if (!id)
    {
        /* Pretend its fastened */
        XMLElement *fasten = XMLookForTKV(
            stanza, "apply-to", 
            "xmlns", "urn:xmpp:fasten:0"
        );
        retract = XMLookForTKV(
            fasten, "retract", 
            "xmlns", "urn:xmpp:message-retract:0"
        );

        if (retract)
        {
            id = HashMapGet(fasten->attrs, "id");
        }
    }

    return id;
}
char *
XMPPGetReply(XMLElement *elem)
{
    XMLElement *rep = XMLookForTKV(elem, "reply", "xmlns", "urn:xmpp:reply:0");
    if (!rep)
    {
        return NULL;
    }

    return HashMapGet(rep->attrs, "id");
}
ssize_t
XMPPGetReplyOffset(XMLElement *elem)
{
    if (!elem)
    {
        return -1;
    }
    for (size_t i = 0; i < ArraySize(elem->children); i++)
    {
        XMLElement *child = ArrayGet(elem->children, i);
        char *xmlns = HashMapGet(child->attrs, "xmlns");
        char *xfor = HashMapGet(child->attrs, "for");
        if (StrEquals(child->name, "fallback") && 
            StrEquals(xmlns, "urn:xmpp:feature-fallback:0") && 
            StrEquals(xfor, "urn:xmpp:reply:0"))
        {
            XMLElement *body = XMLookForUnique(child, "body");
            if (body && HashMapGet(body->attrs, "end"))
            {
                return strtol(HashMapGet(body->attrs, "end"), NULL, 10);
            }
        }
    }
    return -1;
}
char *
XMPPGetModeration(XMLElement *stanza)
{
#define MOD_BASE "urn:xmpp:message-moderate"
    XMLElement *moderate;
    XMLElement *fasten;
    if (!stanza)
    {
        return NULL;
    }

    /* 0.3.0 version of XEP-0425: Bare look at the moderate element */
    moderate = XMLookForTKV(stanza, "moderate", "xmlns", MOD_BASE ":1");
    if (moderate)
    {
        return HashMapGet(moderate->attrs, "id");
    }
    
    /* Earlier versions, which still depend on fastening. */
    fasten = XMLookForTKV(
        stanza, "apply-to", 
        "xmlns", "urn:xmpp:fasten:0"
    );
    moderate = XMLookForTKV(
        fasten, "moderated", 
        "xmlns", MOD_BASE ":0"
    );
    if (moderate)
    {
        return HashMapGet(fasten->attrs, "id");
    }
    return NULL;
#undef MOD_BASE
}

bool 
XMPPHasError(XMLElement *stanza, char *type)
{
    XMLElement *err;
    if (!stanza || !type)
    {
        return false;
    }

    err = XMLookForUnique(stanza, "error");
    if (!err)
    {
        return false;
    }
    return XMLookForUnique(err, type);
}

XMLElement *
XMPPSendDisco(ParseeData *data, char *from, char *to)
{
    XMPPComponent *jabber = data ? data->jabber : NULL;
    XMLElement *ret, *iq;
    char *identifier;

    if (!data || !jabber || !from || !to)
    {
        return NULL;
    }

    pthread_mutex_lock(&data->oidl);
    if ((ret = HashMapGet(data->oid_servers, to)))
    {
        ret = XMLCopy(ret);
        pthread_mutex_unlock(&data->oidl);
        return ret;
    }
    pthread_mutex_unlock(&data->oidl);

    iq = XMLCreateTag("iq");
    XMLAddAttr(iq, "type", "get");
    XMLAddAttr(iq, "from", from);
    XMLAddAttr(iq, "to", to);
    XMLAddAttr(iq, "id", (identifier = StrRandom(64)));
    {
        XMLElement *query = XMLCreateTag("query");
        XMLAddAttr(query, "xmlns", "http://jabber.org/protocol/disco#info");
        XMLAddChild(iq, query);
    }

    XMPPSendStanza(jabber, iq, 10000);
    XMLFreeElement(iq);

    ret = ParseeAwaitStanza(identifier, 1.25 SECONDS);
    Free(identifier);
    if (ret)
    {
        /* TODO: On my own instance, disco replies are _really_ expensive.
         * About 120KB. This is sad. Really sad.  */
        pthread_mutex_lock(&data->oidl);
        XMLFreeElement(HashMapSet(data->oid_servers, to, ret));
        ret = XMLCopy(ret);
        pthread_mutex_unlock(&data->oidl);
    }
    return ret;
}
bool
XMPPDiscoAdvertises(XMLElement *disco, char *var)
{
    XMLElement *query;
    if (!disco || !var)
    {
        return false;
    }

    query = XMLookForTKV(
        disco, 
        "query", "xmlns", 
        "http://jabber.org/protocol/disco#info"
    );
    if (!query)
    {
        return false;
    }

    return !!XMLookForTKV(query, "feature", "var", var);
}

char *
XMPPGetErrtype(XMLElement *stanza)
{
    XMLElement *error = XMLookForUnique(stanza, "error");
    XMLElement *type = NULL;
    if (!error)
    {
        return NULL;
    }

    type = ArrayGet(error->children, 0);
    if (!type)
    {
        return NULL;
    }

    return type->name;
}

char *
XMPPGetErrtext(XMLElement *stanza)
{
    XMLElement *error = XMLookForUnique(stanza, "error");
    XMLElement *text = NULL, *data;
    if (!error)
    {
        return NULL;
    }

    text = XMLookForUnique(error, "text");
    if (!text)
    {
        return NULL;
    }

    data = ArrayGet(text->children, 0);

    return data ? data->data : NULL;

}
