#include <XMPPCommand.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Str.h>

struct XMPPOption {
    enum {
        XMPP_OPTION_TEXT,
        XMPP_OPTION_BOOL,
        XMPP_OPTION_LIST,
        XMPP_OPTION_FIXED
    } type;

    union {
        /* Default text/value for string/fixed */
        struct {
            char *text;
        } as_string;

        /* Default choice for bool */
        struct {
            bool choice;
        } as_bool;

        struct {
            Array *elements;
            char *def;
            bool single;
        } as_arr;
    } data;

    char *id, *desc;
    bool required;
};

void
XMPPFreeOption(XMPPOption *opt)
{
    size_t i;
    Array *elements;
    if (!opt)
    {
        return;
    }

    switch (opt->type)
    {
        case XMPP_OPTION_TEXT:
        case XMPP_OPTION_FIXED:
            Free(opt->data.as_string.text);
            break;
        case XMPP_OPTION_LIST:
            elements = opt->data.as_arr.elements;
            for (i = 0; i < ArraySize(elements); i++)
            {
                char *element = ArrayGet(elements, i);
                Free(element);
            }
            ArrayFree(elements);
            Free(opt->data.as_arr.def);
            break;
        default:
            break;
    }

    Free(opt->desc);
    Free(opt->id);
    Free(opt);
}

XMPPOption *
XMPPCreateText(bool req, char *id, char *def)
{
    XMPPOption *opt;
    if (!id)
    {
        return NULL;
    }
    
    opt = Malloc(sizeof(*opt));
    opt->desc = NULL;
    opt->type = XMPP_OPTION_TEXT;
    opt->id = StrDuplicate(id);
    opt->data.as_string.text = StrDuplicate(def);
    opt->required = req;
    return opt;
}
XMPPOption *
XMPPCreateBool(bool req, char *id, bool def)
{
    XMPPOption *opt;
    if (!id)
    {
        return NULL;
    }
    
    opt = Malloc(sizeof(*opt));
    opt->desc = NULL;
    opt->type = XMPP_OPTION_BOOL;
    opt->id = StrDuplicate(id);
    opt->data.as_bool.choice = def;
    opt->required = req;
    return opt;

}
XMPPOption *
XMPPCreateList(bool req, bool single, char *id, char *def)
{
    XMPPOption *opt;
    if (!id)
    {
        return NULL;
    }
    
    opt = Malloc(sizeof(*opt));
    opt->desc = NULL;
    opt->type = XMPP_OPTION_LIST;
    opt->id = StrDuplicate(id);
    opt->data.as_arr.def = StrDuplicate(def);
    opt->data.as_arr.single = single;
    opt->data.as_arr.elements = ArrayCreate();
    opt->required = req;

    ArrayAdd(opt->data.as_arr.elements, StrDuplicate(def));
    return opt;
}
XMPPOption *
XMPPCreateFixed(char *id, char *text)
{
    XMPPOption *opt;
    if (!id)
    {
        return NULL;
    }
    
    opt = Malloc(sizeof(*opt));
    opt->desc = NULL;
    opt->type = XMPP_OPTION_TEXT;
    opt->id = StrDuplicate(id);
    opt->data.as_string.text = StrDuplicate(text);
    opt->required = false;
    return opt;
}


XMLElement *
XMPPOptionToXML(XMPPOption *opt)
{
    XMLElement *elem, *req, *desc, *data;
    XMLElement *value;
    if (!opt)
    {
        return NULL;
    }

    elem = XMLCreateTag("field");
    switch (opt->type)
    {
        case XMPP_OPTION_TEXT: XMLAddAttr(elem, "type", "text-single"); break;
        case XMPP_OPTION_BOOL: XMLAddAttr(elem, "type", "boolean"); break;

                               /* TODO */
        case XMPP_OPTION_LIST: 
           XMLAddAttr(elem, 
               "type", 
               opt->data.as_arr.single ? "list-single" : "list-multi"
           );
           break;
        case XMPP_OPTION_FIXED: XMLAddAttr(elem, "type", "fixed"); break;
    }
    XMLAddAttr(elem, "var", opt->id);
    if (opt->required)
    {
        req = XMLCreateTag("required");
        XMLAddChild(elem, req);
    }
    if (opt->desc)
    {
        desc = XMLCreateTag("desc");
        data = XMLCreateText(opt->desc);
        XMLAddChild(desc, data);
        XMLAddChild(elem, desc);

        XMLAddAttr(elem, "label", opt->desc);
    }

    if (opt->type == XMPP_OPTION_LIST)
    {
        size_t i;
        Array *elements = opt->data.as_arr.elements;
        XMLElement *option, *value;

        for (i = 0; i < ArraySize(elements); i++)
        {
            char *element = ArrayGet(elements, i);
            option = XMLCreateTag("option");
            value = XMLCreateTag("value");
            data = XMLCreateText(element);

            XMLAddChild(value, data);
            XMLAddChild(option, value);
            XMLAddChild(elem, option);
        }
    }

    switch (opt->type)
    {
        case XMPP_OPTION_TEXT:
            value = XMLCreateTag("value");
            data = XMLCreateText(opt->data.as_string.text);
            XMLAddChild(value, data);
            XMLAddChild(elem, value);
            break;
        case XMPP_OPTION_LIST:
            value = XMLCreateTag("value");
            data = XMLCreateText(opt->data.as_arr.def);
            XMLAddChild(value, data);
            XMLAddChild(elem, value);
            break;
        case XMPP_OPTION_BOOL:
            value = XMLCreateTag("value");
            data = XMLCreateText(opt->data.as_bool.choice ? "true" : "false");
            XMLAddChild(value, data);
            XMLAddChild(elem, value);
            break;
        default:
            break;

    }
    /* TODO */

    return elem;
}


void 
XMPPAddListOption(XMPPOption *list, char *option)
{
    Array *elements;
    if (!list || !option || list->type != XMPP_OPTION_LIST)
    {
        return;
    }
    
    elements = list->data.as_arr.elements;
    ArrayAdd(elements, StrDuplicate(option));
}
void
XMPPSetDescription(XMPPOption *opt, char *desc)
{
    if (!opt || !desc)
    {
        return;
    }

    Free(opt->desc);
    opt->desc = StrDuplicate(desc);
}
bool
XMPPIsOptionRequired(XMPPOption *opt)
{
    return opt ? opt->required : false;
}
char *
XMPPOptionVar(XMPPOption *opt)
{
    if (!opt)
    {
        return NULL;
    }

    return opt->id;
}
