#include "XMPPThread/internal.h"

#include <Cytoplasm/HashMap.h>
#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Str.h>
#include <Cytoplasm/Log.h>

#include <pthread.h>

#include <XMPP.h>
#include <XML.h>

#define PUBSUB "http://jabber.org/protocol/pubsub"

XMLElement *
CreatePubsubRequest(char *from, char *to, char *node)
{
    XMLElement *iq_req, *pubsub, *sub;
    char *id;
    iq_req = XMLCreateTag("iq");
    XMLAddAttr(iq_req, "from", from);
    XMLAddAttr(iq_req, "to", to);
    XMLAddAttr(iq_req, "id", (id = StrRandom(16)));
    XMLAddAttr(iq_req, "type", "set");

    pubsub = XMLCreateTag("pubsub");
    XMLAddAttr(pubsub, "xmlns", PUBSUB);
    XMLAddChild(iq_req, pubsub);

    sub = XMLCreateTag("subscribe");
    XMLAddAttr(sub, "node", node);
    XMLAddAttr(sub, "jid", from);
    XMLAddChild(pubsub, sub);

    Free(id);
    return iq_req;
}

static bool
IsPubsubRequest(XMLElement *stanza)
{
    char *type = HashMapGet(stanza ? stanza->attrs : NULL, "type");
    XMLElement *pubsub;
    if (!stanza)
    {
        return false;
    }

    if (!StrEquals(stanza->name, "iq") ||
        !StrEquals(type, "set"))
    {
        return false;
    }

    pubsub = XMLookForTKV(stanza, "pubsub", "xmlns", PUBSUB);
    if (!pubsub)
    {
        return false;
    }

    return XMLookForUnique(pubsub, "subscribe");
}


struct PEPManager {
    pthread_mutex_t lock;

    ParseeData *data;
    HashMap *node_table;

    HashMap *followers;

    void *cookie;
};

PEPManager *
CreatePEPManager(ParseeData *data, void *cookie)
{
    PEPManager *ret;
    if (!data)
    {
        return NULL;
    }

    ret = Malloc(sizeof(*ret));
    ret->cookie = cookie;
    ret->data = data;
    ret->node_table = HashMapCreate();
    ret->followers = HashMapCreate();
    pthread_mutex_init(&ret->lock, NULL);

    return ret;
}
void *
PEPManagerCookie(PEPManager *manager)
{
    return manager ? manager->cookie : NULL;
}
void
PEPManagerAddEvent(PEPManager *manager, char *node, PEPEvent event)
{
    PEPEvent *indirect;
    if (!manager || !node || !event)
    {
        return;
    }

    indirect = Malloc(sizeof(event));
    *indirect = event;
    pthread_mutex_lock(&manager->lock);
    HashMapSet(manager->node_table, node, indirect);
    pthread_mutex_unlock(&manager->lock);
}

static bool
PEPManagerHandleEvent(PEPManager *manager, XMLElement *stanza)
{
    PEPEvent *call = NULL;
    XMLElement *event = NULL, *ps = NULL, *ev = NULL;
    size_t i;
    if (!manager || !stanza)
    {
        return false;
    }

    if (!(ps = XMLookForTKV(stanza, "pubsub", "xmlns", PUBSUB)) &&
        !(ev = XMLookForTKV(stanza, "event", "xmlns", PUBSUB "#event")))
    {
        return false;
    }
    event = ps ? ps : ev;

    for (i = 0; i < ArraySize(event->children); i++)
    {
        XMLElement *items = ArrayGet(event->children, i);
        char *node = HashMapGet(items->attrs, "node");

        if ((call = HashMapGet(manager->node_table, node)) && *call)
        {
            size_t j;
            /* Use the callback over all items */
            if (ev)
            {
                for (j = 0; j < ArraySize(items->children); j++)
                {
                    (*call)(manager, stanza, ArrayGet(items->children, j));
                }
                return true;
            }

            /* ... or over "items" specifically. */
            (*call)(manager, stanza, items);
            return true;
        }
    }

    return false;
}

bool
PEPManagerHandle(PEPManager *manager, XMLElement *stanza)
{
    if (!manager || !stanza)
    {
        return false;
    }

    /* Check if it is a PEP stanza */
    if (IsPubsubRequest(stanza))
    {
        Log(LOG_DEBUG, "UNIMPLEMENTED PUBSUB SUBSCRIPTION");
        /* TODO */
    }
    if (PEPManagerHandleEvent(manager, stanza))
    {
        return true;
    }

    return false;
}
void
DestroyPEPManager(PEPManager *manager)
{
    char *key;
    PEPEvent *val;
    if (!manager)
    {
        return;
    }

    pthread_mutex_destroy(&manager->lock);
    while (HashMapIterate(manager->node_table, &key, (void **) &val))
    {
        Free(val);
    }
    HashMapFree(manager->node_table);

    HashMapFree(manager->followers);

    Free(manager);
}
