/* The reader-listeners system for the XPPP thread. */
#include <Parsee.h>

#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>

#include <Cytoplasm/Memory.h>
#include <Cytoplasm/Base64.h>
#include <Cytoplasm/Util.h>
#include <Cytoplasm/Log.h>
#include <Cytoplasm/Str.h>

#include <StringStream.h>
#include <XMPPCommand.h>
#include <Matrix.h>
#include <XMPP.h>
#include <XML.h>
#include <AS.h>

#include "XMPPThread/internal.h"

XMLElement *
RetrieveStanza(XMPPThread *thread)
{
    XMLElement *ret = NULL;

    /* TODO: We're polling here... Parsee should avoid doing that,
     * in favour of condition variables, which should lower the system 
     * load of Parsee as a whole. */
    pthread_mutex_lock(&thread->info->lock);
    ret = ArrayDelete(thread->info->stanzas, 0);
    pthread_mutex_unlock(&thread->info->lock);

    return ret;
}
static void
PushStanza(XMPPThreadInfo *info, XMLElement *stanza)
{
    pthread_mutex_lock(&info->lock);
    ArrayAdd(info->stanzas, stanza);
    pthread_mutex_unlock(&info->lock);
}

static void *
XMPPDispatcher(void *argp)
{
    XMPPThread *thread = argp;
    ParseeData *args = thread->info->args;

    while (thread->info->running)
    {
        XMLElement *stanza = RetrieveStanza(thread);
 
        if (!stanza)
        {
            /* TODO: We shouldn't be busywaiting. Even with a sleep call.
             */
            UtilSleepMillis(10);
            continue;
        }

        if (args->verbosity >= PARSEE_VERBOSE_COMICAL)
        {
            Log(LOG_DEBUG, "Received stanza='%s'.", stanza->name);
        } 

        if (ManageMUCStanza(args->muc, stanza))
        {
            XMLFreeElement(stanza);
            continue;
        }

        if (StrEquals(stanza->name, "presence"))
        {
            PresenceStanza(args, stanza, thread);
            XMLFreeElement(stanza);
            continue;
        }
        else if (StrEquals(stanza->name, "message"))
        {
            if (!MessageStanza(args, stanza, thread))
            {
                if (args->verbosity >= PARSEE_VERBOSE_COMICAL)
                {
                    Log(LOG_DEBUG, "Done receiving message with an error");
                } 
                continue;
            }
        }
        else if (StrEquals(stanza->name, "iq"))
        {
            IQStanza(args, stanza, thread);
        }
        else
        {
            Log(LOG_WARNING, "Unknown stanza '%s':", stanza->name);
            XMLEncode(StreamStdout(), stanza);
            StreamPrintf(StreamStdout(), "\n");
            StreamFlush(StreamStdout());
        }

        if (args->verbosity >= PARSEE_VERBOSE_COMICAL)
        {
            Log(LOG_DEBUG, "Done receiving stanza='%s'.", stanza->name);
        } 
        XMLFreeElement(stanza);
    }

    return NULL;
}

typedef struct XMPPAwait {
    XMLElement *stanza;
    bool usable;

    pthread_mutex_t cond_lock;
    pthread_cond_t condition;
} XMPPAwait;

static pthread_mutex_t await_lock = PTHREAD_MUTEX_INITIALIZER;
static HashMap *await_table = NULL;
static XMPPThreadInfo info;

size_t
ParseeCongestion(void)
{
    size_t congestion;
    pthread_mutex_lock(&info.lock);
    congestion = ArraySize(info.stanzas);
    pthread_mutex_unlock(&info.lock);

    return congestion;
}
bool 
XMPPCommandFilter(XMPPCommandManager *m, char *id, XMLElement *stanza)
{
    ParseeData *args = XMPPGetManagerCookie(m);
    char *trimmed_from;
    char *from;
    char *chat_id;
    bool is_muc;
    if (!m || !id || !stanza)
    {
        return false;
    }

    from = HashMapGet(stanza->attrs, "from");
    trimmed_from = ParseeTrimJID(from);
    is_muc = !!(chat_id = ParseeGetFromMUCID(args, trimmed_from));
    Free(trimmed_from);
    Free(chat_id);
#define XMPP_COMMAND(f,l,n,t,s) \
    if (StrEquals(n, id)) \
    { \
        if (l == XMPPCMD_ALL) \
        { \
            return true; \
        } \
        else if (l == XMPPCMD_MUC) \
        { \
            return is_muc; \
        } \
        else if (l == XMPPCMD_ADMINS) \
        { \
            bool is_admin; \
            trimmed_from = ParseeTrimJID(from); \
            is_admin = ParseeIsAdmin(args, trimmed_from); \
            Free(trimmed_from); \
            return is_admin; \
        } \
    }
    XMPPCOMMANDS
#undef XMPP_COMMAND

    return false;
}

void *
ParseeXMPPThread(void *argp)
{
    ParseeData *args = argp;
    XMPPComponent *jabber = args->jabber;
    XMLElement *stanza = NULL;
    HashMap *await_table2;
    size_t i;

    bool error = false;
        
    /* Initialise the await table */
    await_table = HashMapCreate();

    /* Initialise the managers, and add all handlers. */
    info.m = XMPPCreateManager(args);
    XMPPManagerSetFilter(info.m, XMPPCommandFilter);
    {
        XMPPCommand *cmd;
#define XMPP_COMMAND(f,l,n,t,s) \
        cmd =  XMPPBasicCmd( \
            n, t, f \
        ); \
        s \
        XMPPRegisterCommand(info.m, cmd);
        XMPPCOMMANDS
#undef XMPP_COMMAND
    }
    info.pep_manager = CreatePEPManager(args, &info);
    {
        PEPManagerAddEvent(
            info.pep_manager, 
            "urn:xmpp:avatar:metadata", PEPAvatarEvent
        );
        PEPManagerAddEvent(
            info.pep_manager, 
            "urn:xmpp:vcard4", PEPVCardEvent
        );
    }

    /* Initialise the FIFO */
    info.stanzas = ArrayCreate();
    pthread_mutex_init(&info.lock, NULL);

    /* ... and its readers. */
    info.available_dispatchers = args->config->xmpp_threads;
    info.dispatchers = Malloc(
        sizeof(*info.dispatchers) * info.available_dispatchers
    );

    info.args = args;
    info.jabber = jabber;
    info.running = true;
    pthread_mutex_init(&info.chk_lock, NULL);

    for (i = 0; i < info.available_dispatchers; i++)
    {
        pthread_t *thr = &info.dispatchers[i].thr;
        info.dispatchers[i].info = &info;

        if (pthread_create(thr, NULL, XMPPDispatcher, &info.dispatchers[i]))
        {
            Achievement("COULDNT INITIALISE XMPP DISPATCHERS", true);
        }
    }

    while (!args->halted)
    {
        while (true)
        {
            char *id;

            stanza = XMLDecode(jabber->stream, false);
            if (!stanza)
            {
                /* Try to check if an error is abound */
                if (args->verbosity >= PARSEE_VERBOSE_COMICAL)
                {
                    Log(LOG_DEBUG, "RECEIVED EOF.");
                }
                break;
            }

            if (args->verbosity >= PARSEE_VERBOSE_STANZA)
            {
                Stream *output = StreamStderr();
                StreamPrintf(output,   "-------STANZA BEGIN-------"     "\n");
                XMLEncode(output, stanza);
                StreamPrintf(output, "\n--------STANZA END--------"     "\n");
                StreamFlush(output);
            }

            id = HashMapGet(stanza->attrs, "id");
            if (id)
            {
                XMPPAwait *await;
                /* Lock out the table to see if we're awaiting. */
                pthread_mutex_lock(&await_lock);
                if ((await = HashMapGet(await_table, id)))
                {
                    pthread_mutex_lock(&await->cond_lock);
                    await->stanza = stanza;             
                    pthread_cond_signal(&await->condition);
                    pthread_mutex_unlock(&await->cond_lock);

                    HashMapDelete(await_table, id);

                    pthread_mutex_unlock(&await_lock);
                    continue;
                }
                pthread_mutex_unlock(&await_lock);
            }

            /* Push it into the stanza FIFO. A dispatcher thread should then 
             * be able to freely grab a value(locked by a mutex). We can't make 
             * dispatchers read stanzas on their own, since that's _not_ how 
             * streams work, but this should mitigate some issues, and allow a 
             * few threads to be busy, while the rest of Parsee works. */
            PushStanza(&info, stanza);
        }
        pthread_mutex_lock(&args->halt_lock);
        if (!args->halted)
        {
            Log(LOG_WARNING, "XMPP server is closing stream...");
            for (size_t i = 0; i < 50; i++)
            {
                UtilSleepMillis(100); /* Wait a bit so that temporary failures don't fuck everything up */

                Log(LOG_WARNING, "Restarting XMPP stream.");
                /* This is the part where a new connection is being considered */
                XMPPFinishCompStream(jabber);
                XMPPEndCompStream(jabber);

                args->jabber = XMPPInitialiseCompStream(
                    args->config->component_addr, 
                    args->config->component_host, 
                    args->config->component_port
                );
                jabber = args->jabber;
                if (!jabber || !XMPPAuthenticateCompStream(jabber, args->config->shared_comp_secret))
                {
                    /* Oops, there is something wrong! */
                    Log(LOG_ERR, "Couldn't authenticate to XMPP server");
                    XMPPEndCompStream(jabber);
                    args->jabber = NULL;
                    jabber = NULL;
                    if (i == 4)
                    {
                        pthread_mutex_unlock(&args->halt_lock);
                        break;
                    }
                }
            }
        }
        pthread_mutex_unlock(&args->halt_lock);
    }

    info.running = false;
    for (i = 0; i < info.available_dispatchers; i++)
    {
        pthread_t thr = info.dispatchers[i].thr;
        if (pthread_join(thr, NULL))
        {
            Achievement("COULDNT JOIN XMPP DISPATCHER", true);
        }

        Log(LOG_DEBUG, "- joined %d/%d...", 
            i + 1, info.available_dispatchers
        );
    }
    Log(LOG_INFO, "Joined subthreads...");
    Free(info.dispatchers);

    if (ArraySize(info.stanzas))
    {
        Log(LOG_DEBUG, "(%d unprocessed stanzas)", ArraySize(info.stanzas));
    }
    for (i = 0; i < ArraySize(info.stanzas); i++)
    {
        XMLFreeElement(ArrayGet(info.stanzas, i));
    }
    ArrayFree(info.stanzas);

    await_table2 = await_table;
    await_table = NULL;
    HashMapFree(await_table2);

    pthread_mutex_destroy(&info.lock);

    DestroyPEPManager(info.pep_manager);
    XMPPFreeManager(info.m);
    if (error)
    {
        HttpServerStop(args->server);   /* minor trolling */
    }
    return NULL;
}


XMLElement *
ParseeAwaitStanza(char *identifier, int64_t timeout)
{
    /* XXX: You may be wondering _why_ I am using stdlib's malloc, instead of 
     * Cytoplasm's Malloc wrapper(which are essentially the same thing, plus 
     * some memchecking as of now), or the stack, as done previously.
     * The reasons are that:
     *  - Cytoplasm's heap causes a fun bus error on ARM64. I have _no clue_ why, 
     *  since, as I said, it is as of now just malloc with extra metadata prepended 
     *  to it. A _guess_ may be that the metadata makes everything unaligned, which, 
     *  just sounds silly. pthreads should be able to cope fine with that, I think.
     *
     *  I am open to issues talking about such fun behaviour and ideas to fix this. 
     *  This is a bodge, and potentially a problem, or vulnerability with Cytoplasm/
     *  Parsee. As long as it isn't "use Rust". Please don't say that. */
    XMPPAwait awa = { 0 };
    XMLElement *stanza;
    struct timespec ts;

    if (!identifier)
    {
        return NULL;
    }

    pthread_mutex_lock(&await_lock);
    if (HashMapGet(await_table, identifier))
    {
        return NULL;
    }
    pthread_mutex_unlock(&await_lock);


    /* Convert into an absolute timeout.
     * =================================
     * XXX: For anyone using timespecs: MAKE ABSOLUTELY FUCKING SURE YOUR NANOS 
     * ARE IN RANGE. THIS UNIRONICALLY CAUSED SOME REAL CONCURRENCY ERRORS ON MY 
     * SIDE FROM TV_NSEC OVERFLOWING AND TIMEDWAIT RETURNING EINVAL. */
    if (timeout > 0)
    {
        int64_t seconds = timeout / (1 SECONDS);
        int64_t msecond = timeout % (1 SECONDS);

        clock_gettime(CLOCK_REALTIME, &ts);

        ts.tv_sec += seconds            ;
        ts.tv_nsec+= msecond * 1000000  ;
        if (ts.tv_nsec > 999999999)
        {
            int64_t sec_delta = ts.tv_nsec / 1000000000;
            int64_t nsc_delta = ts.tv_nsec % 1000000000;

            ts.tv_nsec = nsc_delta;
            ts.tv_sec += sec_delta;
        }
    }

    pthread_mutex_lock(&await_lock);

    pthread_cond_init(&awa.condition, NULL);
    pthread_mutex_init(&awa.cond_lock, NULL);
    awa.stanza = NULL;

    HashMapSet(await_table, identifier, &awa);
    pthread_mutex_unlock(&await_lock);

    pthread_mutex_lock(&awa.cond_lock);
    while (!awa.stanza)
    {
        int code;

        if (timeout <= 0)
        {
            pthread_cond_wait(&awa.condition, &awa.cond_lock);
            continue;
        }
        code = pthread_cond_timedwait(&awa.condition, &awa.cond_lock, &ts);
        if (code == ETIMEDOUT)
        {
            /* Timeout detected, give up regardless of the status of our 
             * search. */
            pthread_mutex_lock(&await_lock);
            HashMapDelete(await_table, identifier);
            pthread_mutex_unlock(&await_lock);
            break;
        }
        if (code == EINVAL)
        {
            Achievement("TIMEDWAIT RETURNED WEIRD EINVAL", true);
        }
    }

    stanza = awa.stanza;
    pthread_mutex_unlock(&awa.cond_lock);

    pthread_mutex_lock(&await_lock);
    pthread_cond_destroy(&awa.condition);
    pthread_mutex_destroy(&awa.cond_lock);
    pthread_mutex_unlock(&await_lock);
    return stanza;
}
