#ifndef PARSEE_XMPPCOMMAND_H
#define PARSEE_XMPPCOMMAND_H

#include <Cytoplasm/HashMap.h>
#include <stdbool.h>

#include <Parsee.h>
#include <XML.h>

typedef struct XMPPCommandManager XMPPCommandManager;
typedef struct XMPPCommand XMPPCommand;
typedef struct XMPPOption XMPPOption;
typedef void (*XMPPCmdCallback)(XMPPCommandManager *, char *, XMLElement *, XMLElement *);
typedef void (*XMPPOptionWriter)(XMPPCommandManager *, XMPPCommand *, char *);
typedef bool (*XMPPCmdFilter)(XMPPCommandManager *, char *id, XMLElement *stanza); 

/** Creates a simple XMPP command manager, which routes commands 
 * with a single-stage form system, with a {cookie}
 * -------------------------------------------
 * Returns: An opaque command manager[LA:HEAP] 
 * Modifies: NOTHING
 * See-Also: XMPPFreeManager */
extern XMPPCommandManager * XMPPCreateManager(void *cookie); 
extern void * XMPPGetManagerCookie(XMPPCommandManager *manager);

/** Changes the filter used for the command manager. This function 
 * takes in the source stanza and the command "ID" to filter, and 
 * returns true IFF the command should be shown to the requester.
 * -----------
 * Returns: NOTHING
 * Modifies: the manager's filter
 * See-Also: XMPPCreateManager */
extern void
XMPPManagerSetFilter(XMPPCommandManager *manager, XMPPCmdFilter filter);

/** Create a basic command with a node and name description
 * -----------------------------------------------------
 * Returns: A command to be used with {XMPPRegisterCommand}[LA:HEAP]
 * Modifies: NOTHING
 * See-Also: XMPPRegisterCommand */
extern XMPPCommand * 
 XMPPBasicCmd(char *node, char *name, XMPPCmdCallback cb);
extern void 
XMPPCmdOptionsCreator(XMPPCommand *cmd, XMPPOptionWriter writer);
extern void XMPPAddOption(XMPPCommand *cmd, XMPPOption *opt);
extern XMLElement * XMPPFormifyCommand(XMPPCommandManager *m, XMPPCommand *cmd, char *from);
extern char * XMPPGetCommandNode(XMPPCommand *cmd);
extern char * XMPPGetCommandDesc(XMPPCommand *cmd);
extern void  XMPPSetFormInstruction(XMPPCommand *cmd, char *instruction);
extern void  XMPPSetFormTitle(XMPPCommand *cmd, char *title);
extern bool XMPPCommandRequiresForm(XMPPCommand *cmd);
extern void XMPPExecuteCommand(XMPPCommandManager *m, XMPPCommand *cmd, char *from, XMLElement *to, XMLElement *form);

/** Create a basic option.
 * -----------------------------------------------------
 * Returns: A form option of the right type[LA:HEAP]
 * Modifies: NOTHING
 * See-Also: https://xmpp.org/extensions/xep-0004.html */
extern XMPPOption * XMPPCreateText(bool req, char *id, char *def);
extern XMPPOption * XMPPCreateBool(bool req, char *id, bool def);
extern XMPPOption * XMPPCreateList(bool req, bool single, char *id, char *def);
extern XMPPOption * XMPPCreateFixed(char *id, char *text);

extern void XMPPAddListOption(XMPPOption *list, char *option);
extern void XMPPSetDescription(XMPPOption *opt, char *descri);

extern bool XMPPIsOptionRequired(XMPPOption *opt);
extern char * XMPPOptionVar(XMPPOption *opt);

extern XMLElement * XMPPOptionToXML(XMPPOption *opt);

extern void XMPPFreeOption(XMPPOption *opt);
extern void XMPPFreeCommand(XMPPCommand *cmd);

extern bool XMPPVerifyForm(XMPPCommand *cmd, XMLElement *form);

/** Registers a {cmd} to the {m}anager, with no extra metadata or callback. 
 * It also makes {cmd} belong to {m}anager, and therefore does not belong to 
 * its creator anymore.
 * -----------------------------------------------------
 * Returns: NOTHING
 * Modifies: {m}anager
 * See-Also: XMPPCreateManager, XMPPFreeManager */
extern void XMPPRegisterCommand(XMPPCommandManager *m, XMPPCommand *cmd);

/** Shoves all {m} commands into XML as children of {p}, and a {jid}
 * (and with the stanza source)
 * -----------------------------------------------------
 * Returns: NOTHING
 * Modifies: {p}
 * See-Also: XMPPCreateManager */
extern void XMPPShoveCommandList(XMPPCommandManager *m, char *jid, XMLElement *p, XMLElement *s);

/** Destroys all memory related to the command {manager}.
 * -----------------------------------------------------
 * Returns: NOTHING
 * Modifies: {manager}
 * See-Also: XMPPCreateManager */
extern void XMPPFreeManager(XMPPCommandManager *manager);

/** Manages a command given a IQ stanza and a Parsee data field
 * -----------------------------------------------------
 * Returns: Whenever the command was managed properly and completed.
 * Modifies: {manager}
 * See-Also: XMPPCreateManager */
extern bool XMPPManageCommand(XMPPCommandManager *m, XMLElement *stanza, ParseeData *data);


/* --------------------------------- COMMANDS --------------------------------- */
/* Please edit stc/XMPPThread.c (you can just force-save) for these to apply! */

#define XMPP_COMMAND(f,l,n,t,s) \
    extern void \
    f(XMPPCommandManager *, char *, XMLElement *, XMLElement *); \
    /* This symbol might not exist. We do, however, not care, as 
     * it is not always done. */ \
    extern void \
    Form##f(XMPPCommandManager *, XMPPCommand *, char *); \

#include "XMPPCommands.x.h"

#undef XMPP_COMMAND
#endif
